// Use bitfields instead of HashSets.
use bit_field::BitField;

// If I wanted to do 4x4 or 5x5, I would need more bits. (u32).

use num::Integer;
use std::ops::RangeBounds;

#[derive(Debug, Copy, Clone, PartialEq)]
pub struct GenBits<T: Integer + BitField>(pub T);

impl<T: Integer + BitField> GenBits<T> {
    pub fn clear(&mut self) {
        self.0 = T::zero();
    }

    pub fn set(&mut self, bit: usize, value: bool) {
        self.0.set_bit(bit, value);
    }

    pub fn get(&self, bit: usize) -> bool {
        self.0.get_bit(bit)
    }

    pub fn set_bits<R: RangeBounds<u8> + IntoIterator<Item = u8>>(&mut self, bits: R) {
        for i in bits {
            self.set(i as usize, true);
        }
    }

    pub fn count_set(&self) -> u8 {
        let mut count: u8 = 0;
        for i in 0..T::BIT_LENGTH {
            if self.get(i) {
                count += 1;
            }
        }
        count
    }
}

pub struct GenBitsIterator<'a, T: Integer + BitField> {
    pub possible: &'a GenBits<T>,
    pub index: usize,
}

impl<T: Integer + BitField> GenBits<T> {
    pub fn iter(&self) -> GenBitsIterator<T> {
        GenBitsIterator {
            possible: self,
            index: 1,
        }
    }
}

impl<'a, T: Integer + BitField> Iterator for GenBitsIterator<'a, T> {
    type Item = u8;

    fn next(&mut self) -> Option<u8> {
        while (self.index < T::BIT_LENGTH) && (!self.possible.get(self.index as usize)) {
            self.index += 1;
        }
        if self.index == T::BIT_LENGTH {
            None
        } else {
            self.index += 1;
            Some((self.index - 1) as u8)
        }
    }
}

#[derive(Debug, Copy, Clone, PartialEq)]
pub struct Bits(pub u16);

/// Set bits number of bits to 1 (true)
pub const fn set_bits(bits: u8) -> u16 {
    (1 << (bits)) - 1
}

impl Bits {
    /// clear all bits
    pub fn clear(&mut self) {
        self.0 = 0;
    }

    /// set bit to state of value.
    pub fn set(&mut self, bit: u8, value: bool) {
        self.0.set_bit(bit as usize, value);
    }

    /// get state of bit.
    pub fn get(&self, bit: u8) -> bool {
        self.0.get_bit(bit as usize)
    }

    /// set bits on, given number of bits initially to set.
    pub fn set_bits(&mut self, bits: u8) {
        self.0 = set_bits(bits);
    }

    /// count number of bits set.
    pub fn count_set(&self) -> u8 {
        let mut count = 0;
        for i in 0..u16::BIT_LENGTH {
            if self.get(i as u8) {
                count += 1;
            }
        }
        count
    }
}

pub struct BitsIterator<'a> {
    pub possible: &'a Bits,
    pub index: u8,
}

impl Bits {
    pub fn iter(&self) -> BitsIterator {
        BitsIterator {
            possible: self,
            index: 1,
        }
    }
}

impl<'a> Iterator for BitsIterator<'a> {
    type Item = u8;

    fn next(&mut self) -> Option<u8> {
        while (self.index < u16::BIT_LENGTH as u8) && (!self.possible.get(self.index)) {
            self.index += 1;
            // println!("index = {}", self.index);
        }
        if self.index == u16::BIT_LENGTH as u8 {
            None
        } else {
            self.index += 1;
            Some(self.index - 1)
        }
    }
}

#[cfg(test)]
mod tests {
    // use crate::sudoku::*;
    use crate::bits::*;

    #[test]
    fn check_possible_bitset() {
        let mut p = Bits(0);

        p.clear();

        for i in 0..9 {
            let mut result = p.get(i);
            assert_eq!(result, false);
            p.set(i, true);
            result = p.get(i);
            assert_eq!(result, true);
        }
    }

    #[test]
    fn check_possible_iter() {
        let mut p = Bits(0);
        p.set(3, true);
        p.set(5, true);
        p.set(6, true);
        assert_eq!(3, p.count_set());
        let values: Vec<u8> = p.iter().collect();
        assert_eq!(values, vec!(3, 5, 6));
        assert_eq!(3, p.count_set());
        p.set(0, true);
        assert_eq!(4, p.count_set());
    }

    #[test]
    fn check_bits() {
        // Set bits 0-5 (6 bits total)
        let p = Bits(set_bits(6));

        for i in 0..6 {
            let result = p.get(i);
            assert_eq!(result, true);
        }
        assert_eq!(p.get(6), false);
    }

    #[test]
    fn check_u16() {
        let mut p = GenBits::<u16>(0);

        p.clear();

        for i in 0..9 {
            let mut result = p.get(i);
            assert_eq!(result, false);
            p.set(i, true);
            result = p.get(i);
            assert_eq!(result, true);
        }

        p = GenBits::<u16>(0);
        p.set(3, true);
        p.set(5, true);
        p.set(6, true);
        assert_eq!(3, p.count_set());
        let values: Vec<u8> = p.iter().collect();
        assert_eq!(values, vec!(3, 5, 6));
        assert_eq!(3, p.count_set());
        p.set(0, true);
        assert_eq!(4, p.count_set());

        p = GenBits::<u16>(0);
        p.set_bits(0..6);

        for i in 0..6 {
            let result = p.get(i);
            assert_eq!(result, true);
        }
        assert_eq!(p.get(6), false);
    }

    #[test]
    fn check_u32() {
        let mut p = GenBits::<u32>(0);

        p.clear();

        for i in 0..29 {
            let mut result = p.get(i);
            assert_eq!(result, false);
            p.set(i, true);
            result = p.get(i);
            assert_eq!(result, true);
        }

        p = GenBits::<u32>(0);
        p.set(13, true);
        p.set(15, true);
        p.set(26, true);
        assert_eq!(3, p.count_set());
        let values: Vec<u8> = p.iter().collect();
        assert_eq!(values, vec!(13, 15, 26));
        assert_eq!(3, p.count_set());
        p.set(0, true);
        assert_eq!(4, p.count_set());

        p = GenBits::<u32>(0);
        p.set_bits(10..26);

        for i in 10..26 {
            let result = p.get(i);
            assert_eq!(result, true);
        }
        assert_eq!(p.get(9), false);
        assert_eq!(p.get(27), false);
    }

}
