// Read a Ksudoku's XML save file.
use std::collections::HashMap;
use std::error;
// use std::fmt;
use std::fs::File;
use std::io::BufReader;
use xml::reader::{EventReader, XmlEvent};
use xml::writer::XmlEvent as WrXmlEvent;
use xml::EmitterConfig;

/*
#[derive(Debug, Clone)]
struct UnsupportedGroup {
    message: String,
}

impl fmt::Display for UnsupportedGroup {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "Unsupported group: {}", self.message)
    }
}

impl error::Error for UnsupportedGroup {}
*/

/*
What I want from load_ksudoku:

  - Puzzle string
  - Solution string (so I can validate it/check my code)
  - specific-type & type (combined as a string is ok - I think)
  - order (as a number)

I get values (the puzzle), I also need solution.

Having order will allow me to use AnyBoard::new(order) and load it up.  

Currently, it loads, and will fail if it isn't Plain, sudoku, 9.
I'm not sure what future ksudoku puzzles I want to support, so I'd like to
have that as flexible as possible.

save_ksudoku, would need to save those fields as well...
 */

 #[derive(Debug)]
 pub struct Ksudoku {
    /// puzzle string:
    /// _ is blank, starts with 'b' (for 1), 'c' for 2.
    pub puzzle :String,
    /// This contains the solution to the puzzle.
    pub solution: String,
    /// "Plain"
    pub specific_type: String,
    /// "sudoku"
    pub puzzle_type: String,
    /// Puzzle width (9, 16, 25).
    pub order: u8,
 }

 impl Default for Ksudoku {
    fn default() -> Self {
        Ksudoku {
            puzzle: String::new(),
            solution: String::new(),
            specific_type: String::from("Plain"),
            puzzle_type: String::from("sudoku"),
            order: 0,
        }
    }
}

 /// Load ksudoku file, returning Ksudoku structure.
 /// - It is up to the caller to determine if they can handle the types
 ///   and size (order).
pub fn load_ksudoku(filename: std::path::PathBuf) -> Result<Ksudoku, Box<dyn error::Error>> {
    let fh = File::open(filename)?;
    let buffer = BufReader::new(fh);
    let parser = EventReader::new(buffer);
    let mut element_name: String = String::new();    
    let mut details = Ksudoku::default();

    for e in parser {
        match e {
            Ok(XmlEvent::StartElement {
                name, attributes, ..
            }) => {
                element_name = name.local_name.clone();
                if element_name == "graph" {
                    // Check the attributes here
                    // <graph specific-type="Plain" type="sudoku" order="9"/>
                    // {"specific-type": "Plain", "type": "sudoku", "order": "9"}

                    let mut attrs: HashMap<String, String> = HashMap::new();

                    for a in attributes {
                        attrs.insert(a.name.to_string(), a.value);
                    }

                    let blank = String::new();

                    details.specific_type = attrs.get(&"specific-type".to_string()).unwrap_or(&blank).clone();
                    details.puzzle_type = attrs.get(&"type".to_string()).unwrap_or(&blank).clone();
                    let order  = attrs.get(&"order".to_string()).unwrap_or(&blank).clone();

                    if !order.is_empty() {
                        let r = order.parse::<u8>();
                        if r.is_ok() {
                            details.order = r.unwrap();
                        } else {
                            // Failed to convert order to u8...
                            return Err(Box::new(r.unwrap_err()));
                        }
                    }
                }
            }
            Ok(XmlEvent::Characters(text)) => {
                if element_name == "values" {
                    details.puzzle = text.clone();
                }
                if element_name == "solution" {
                    details.solution = text.clone();
                }
            }
            Err(e) => {
                return Err(Box::new(e));
            }
            _ => {}
        }
    }

    Ok(details)
}

/// Write out a ksudoku puzzle file
pub fn save_ksudoku(
    filename: std::path::PathBuf,
    data: &Ksudoku,
) -> Result<(), Box<dyn error::Error>> {
    let fh = File::create(filename)?;
    let mut writer = EmitterConfig::new().perform_indent(true).create_writer(fh);

    let mut event: WrXmlEvent = WrXmlEvent::start_element("ksudoku").into();
    writer.write(event)?;
    event = WrXmlEvent::start_element("game")
        .attr("had-help", "0")
        .attr("msecs-elapsed", "0")
        .into();
    writer.write(event)?;
    event = WrXmlEvent::start_element("puzzle").into();
    writer.write(event)?;

    // Convert order to a string.
    let order = data.order.to_string();

    event = WrXmlEvent::start_element("graph")
        .attr("specific-type", &data.specific_type)
        .attr("type", &data.puzzle_type)
        .attr("order", &order )
        .into();
    writer.write(event)?;
    event = WrXmlEvent::end_element().into();
    writer.write(event)?;

    // values (puzzle)
    event = WrXmlEvent::start_element("values").into();
    writer.write(event)?;
    event = WrXmlEvent::characters(&data.puzzle).into();
    writer.write(event)?;
    event = WrXmlEvent::end_element().into();
    writer.write(event)?;

    // solution
    event = WrXmlEvent::start_element("solution").into();
    writer.write(event)?;
    event = WrXmlEvent::characters(&data.solution).into();
    writer.write(event)?;
    event = WrXmlEvent::end_element().into();
    writer.write(event)?;

    // Apparently, the events are consumed...

    event = WrXmlEvent::end_element().into();
    writer.write(event)?;
    event = WrXmlEvent::end_element().into();
    writer.write(event)?;
    event = WrXmlEvent::end_element().into();
    writer.write(event)?;

    Ok(())
}

#[cfg(test)]

mod tests {
    use crate::ksudoku::*;
    use crate::sudoku::*;
    use std::path::PathBuf;

    fn test_files_path() -> PathBuf {
        
    let mut d = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
    d.push("test_files");
        d
    }
    
    #[test]
    fn load_puzzles() {
        let mut testfile = test_files_path();
        testfile.push("9-puzzle");

        let mut result = load_ksudoku(testfile);
        assert!(result.is_ok());
        let result = result.unwrap();
        assert!(result.puzzle_type == "sudoku");
        assert!(result.specific_type == "Plain");

        let size = (result.order as f64).sqrt() as u8;
        let mut board = AnyBoard::new(size);
        let mut load = board.load_from_tld('b', '_', &result.puzzle);
        assert!(load.is_ok());
        
        // Ok, puzzle is loaded!
        board.display();

        let mut testfile = test_files_path();
        testfile.push("16-puzzle");

        let mut result = load_ksudoku(testfile);
        assert!(result.is_ok());
        let result = result.unwrap();
        assert!(result.puzzle_type == "sudoku");
        assert!(result.specific_type == "Plain");

        let size = (result.order as f64).sqrt() as u8;
        let mut board = AnyBoard::new(size);
        let mut load = board.load_from_tld('b', '_', &result.puzzle);
        assert!(load.is_ok());
        
        // Ok, puzzle is loaded!
        board.display();
        
        let mut testfile = test_files_path();
        testfile.push("25-puzzle");

        let mut result = load_ksudoku(testfile);
        assert!(result.is_ok());
        let result = result.unwrap();
        assert!(result.puzzle_type == "sudoku");
        assert!(result.specific_type == "Plain");

        let size = (result.order as f64).sqrt() as u8;
        let mut board = AnyBoard::new(size);
        let mut load = board.load_from_tld('b', '_', &result.puzzle);
        assert!(load.is_ok());
        
        // Ok, puzzle is loaded!
        board.display();

    }
}